const config = require('./config');
const {attach} = require('node-firebird');
const {promisify} = require('util');
const path = require('path');
const { Order, OrangeData } = require(path.join(__dirname, '.', 'lib'));
const crypto = require('crypto');
const { OrangeDataError } = require(path.join(__dirname, '.', 'lib/errors'));

const fbAttach = promisify(attach);
let agent;
let isRun = false;

async function readBlob(callback) {
    return new Promise((resolve, reject) => {
        if ( (callback) && ({}.toString.call(callback) === '[object Function]') ) {
            callback((err, name, e) => {
                if (err) reject(err)
                else {
                    let j = '';
                    e.on('data', function (chunk) {
                        j = j + chunk;
                    });
                    e.on('end', function () {
                        resolve(j)
                    });
                }
            })
        }
        else resolve('');
    })
}

async function checkReply(q) {
    // выбираем чеки, которым нужно актуализировать статус отправки (успешно чек проведен или нет)
    const row = await q('select FISCAL_DATA_ID, SEND_DATA from CAT_FISCAL_DATA where STATUS = 2 ' + // status=2 - обрабатывается
        'and SYSTEM_TYPE = 1');
    logger.debug('Начинаем проверять статусы чеков');
    for (const rec of row) {
        // считываем blob
        const j_text = await readBlob(rec[1]);
        const j = JSON.parse(j_text);
        // help function to check order's status
        logger.debug('Получение статуса чека operation_id='+j.order.operation_id+', inn='+j.order.inn+', id='+j.order.id);
        const status = await agent.getOrderStatus(j.order.inn, j.order.id);
        if (status) {
            logger.info('Чек (operation_id='+j.order.operation_id+', inn='+j.order.inn+', id='+j.order.id+') успешно пробит онлайн-кассой');
            logger.debug('Информация о пробитом чеке:');
            logger.debug(JSON.stringify(status));
            await q('update CAT_FISCAL_DATA set RECEIVE_DATA=?, STATUS=3 where FISCAL_DATA_ID=?', [ status, rec[0] ]);
        }
        else {
            logger.debug('Чек создан и добавлен в очередь на обработку, но еще не обработан');
        }
    }
}

async function checkSend() {
    // выбираем чеки, которые нужно отправить, потом вызываем checkReply
    // подключаемся к БД
    const options = config.db;
    options.pageSize = 4096;
    const db = await fbAttach(options);
    try {
        const q = promisify(db.execute);
        // выбираем неотпраленные чеки (STATUS = 1) и временно неуспешно отправленные (4), относящиеся к OrangeData (SYSTEM_TYPE = 1)
        const row = await q('select FISCAL_DATA_ID, SEND_DATA from CAT_FISCAL_DATA where STATUS in (1, 4) ' +
          'and SYSTEM_TYPE = 1');
        logger.debug('Начинаем отправлять чеки для регистрации')
        for (const rec of row) {
            // считываем blob
            const j_text = await readBlob(rec[1]);
            logger.debug('Считали из базы данные для отправки:')
            logger.debug(j_text)
            if (j_text == '') {
                logger.warn('Пустые данные для отправки!');
                continue;
            }
            let j;
            try {
                j = JSON.parse(j_text);
            } catch (err) {
                logger.error('В базе для отправки хранится невалидный JSON, ошибка:'+err);
                logger.error('Данные:'+j_text);
            };
            j.order.id = crypto.randomBytes(18).toString('hex');
            j.order.inn = config.organization.inn;
            const order = new Order({
                id: j.order.id,
                inn: j.order.inn,
                key: j.order.inn,
                type: j.order.type,
                customerContact: j.order.customerContact,
                taxationSystem: config.organization.taxationSystem
            });
            j.positions.forEach(position => {
                order.addPosition(position)
            });
            j.payments.forEach(payment => {
                order.addPayment(payment)
            });
            logger.debug('Сохраняем в БД данные о чеке:')
            logger.debug(JSON.stringify(j));
            await q('update CAT_FISCAL_DATA set SEND_DATA=? where FISCAL_DATA_ID=?', [ j, rec[0] ]);
            logger.info('Информация о чеке в БД обновлена, operation_id='+j.order.operation_id);
            logger.info('Отправка чека онлайн-кассе, operation_id='+j.order.operation_id);
            logger.debug('Отправляемые данные: ');
            logger.debug('id: '+JSON.stringify(order.id));
            logger.debug('inn: '+JSON.stringify(order.inn));
            logger.debug('content: '+JSON.stringify(order.content));
            try {
                await agent.sendOrder(order);
                logger.info('Чек успешно отправлен');
                await q('update CAT_FISCAL_DATA set STATUS=2 where FISCAL_DATA_ID=?', [ rec[0] ]);
            } catch (error) {
                if (error instanceof OrangeDataError) {
                    logger.error('При обработке чека operation_id='+j.order.operation_id+' возникла ошибка OrangeData, ' +
                        'она содержит дополнительные сведения:');
                    logger.error(JSON.stringify(error));
                }
                else {
                    logger.error('Ошибка при отправке чека:');
                    logger.error(JSON.stringify(error));
                }
                //STATUS=4 - отправка временно неуспешна (потом повторим)
                await q('update CAT_FISCAL_DATA set STATUS=4, RECEIVE_DATA=? where FISCAL_DATA_ID=?', [ error, rec[0] ]);
            }
        }
        try {
            await checkReply(q);
        } catch (err) {
            logger.error('Ошибка при обработке статусов чеков:')
            logger.error(err);
        }
    } finally {
        db.detach();
    }
}

function startCheck() {
    const retry = () => {
        setTimeout(() => {
            startCheck();
        }, 1000 * 60) // через минуту новая проверка
    }
    logger.info('Выполняем обработку чеков');
    checkSend().then(() => {
        logger.info('Обработка чеков успешно завершена');
        retry();
    }, (err) => {
        logger.error('Ошибка при обработке чеков:');
        logger.error(err);
        retry();
    });
}

function tryCheck(a) {
    if (!a) throw 'OrangeData agent is undefined';
    if (isRun) return;
    isRun = true;
    agent = a;
    startCheck();
}

module.exports.tryCheck = tryCheck;