// логирование
try {
  const config = require('./config');
  const winston = require('winston');
  global.logger = winston.createLogger({
    level: config.log.level,
    exitOnError: false,
    format: winston.format.combine(
        winston.format.timestamp({
          format: 'YYYY-MM-DD HH:mm:ss'
        }),
        winston.format.printf(info =>
          `${info.timestamp} ${info.level}: ${info.message}`+(info.splat!==undefined?`${info.splat}`:" "))
    ),
    transports: [
      new winston.transports.File({ filename: config.log.path, maxsize: config.log.maxSize, maxFiles: config.log.maxFiles,
        handleExceptions: true }),
      new winston.transports.Console({handleExceptions: true})
    ]
  });
} catch (error) {
  console.log('Error initialize log:');
  throw new Error(error)
}
// модули
const config = require('./config');
const fs = require('fs');
const path = require('path');
const fb = require('node-firebird');
const {promisify} = require('util');
const {tryCheck} = require('./check');
const { OrangeData } = require(path.join(__dirname, '.', 'lib'));
const {processingCmd} = require('./processing-cmd');

const cert = fs.readFileSync(path.join(__dirname, './keys/client.crt'));
const key = fs.readFileSync(path.join(__dirname, './keys/client.key'));
const passphrase = config.keyPassphrase;
const ca = fs.readFileSync(path.join(__dirname, './keys/cacert.pem'));
const privateKey = fs.readFileSync(path.join(__dirname, './keys/private_key.pem'));
const apiUrl = undefined; // перекрыто в config 'https://apip.orangedata.ru:2443/api/v2';
const fbAttach = promisify(fb.attach);

const agent = new OrangeData({ apiUrl, cert, key, passphrase, ca, privateKey });
const { createValidator } = require('./lib/validation');
const { validateRequest } = require('./lib/validation/schemes');
const validate = createValidator(validateRequest);

async function processingRequest(j, res) {
  if (j.cmd) {
      // это какая то команда, а не входящий чек
      return await processingCmd(j.cmd, res);
  }
  const v = validate(j);
  if (JSON.stringify(v)!=='{}') {
    res.write(JSON.stringify({response: 'error', error: "Переданный JSON не прошел валидацию, подробности в поле " +
      "errorObj",
      errorObj: v}));
    logger.debug('Переданный JSON не прошел валидацию:')
    logger.debug(JSON.stringify(v))
    return ;
  };
  const options = config.db;
  options.pageSize = 4096;
  const db = await fbAttach(options);
  try {
    await promisify(db.execute)('insert into CAT_FISCAL_DATA (OPERATION_ID, STATUS, SYSTEM_TYPE, SEND_DATA) ' +
        'values (?, 1, 1, ?)',
        [j.order.operation_id, j]);
    res.write(JSON.stringify({response: 'ok'}));
    logger.info('Обработан входящий чек')
  } finally {
    db.detach();
  }
}

// main
async function tryMain() {
  if ( (!Number.isInteger(config.organization.taxationSystem)) || (config.organization.taxationSystem < 0)
      || (config.organization.taxationSystem > 5) ) {
    throw 'Параметр config.organization.taxationSystem указан неверно';
  }
  let isServerStart = false;
  const http = require('http');
  const serv = http.createServer();
  serv.on('error', function(e) {
    logger.error('Ошибка HTTP:');
    logger.error(e);
    if (!isServerStart) process.exit(-1);
  });
  serv.on('request', function(req, res) {
    res.setHeader('Content-type', 'application/json');
    if (req.method === 'POST') {
      logger.debug('Получен POST-запрос');
      let body = '';
      req.on('data', chunk => {
        body += chunk.toString(); // convert Buffer to string
      });
      req.on('end', () => {
        logger.debug('Запрос полностью принят:')
        logger.debug(body);
        let j;
        try { j = JSON.parse(body) }
        catch (e) {};
        if (j) {
          logger.debug('Начинаем выполнение запроса')
          processingRequest(j, res).then(() => {
            res.end();
            logger.debug('Запрос обработан');
          }, (reject) => {
            logger.error('Ошибка при выполнении запроса:');
            logger.error(reject);
            res.writeHead(500);
            res.write(JSON.stringify({response: 'error', error: "Внутренняя ошибка сервера: "+reject}));
            res.end();

          });
        }
        else {
          logger.debug('Запрос не является валидным JSON');
          res.writeHead(400);
          res.write(JSON.stringify({response: 'error', error: 'В запросе ожидаются JSON-данные'}));
          res.end();
        }
      });
    }
    else {
      logger.debug('Пришел не поддерживаемый тип запроса: '+req.method);
      res.writeHead(405);
      res.write(JSON.stringify({response: 'error', error: "Поддерживается только метод POST"}));
      res.end();
    }
  });
  serv.listen(config.openHttpPort, function(e) {
    if (e) {
      logger.error('Ошибка HTTP при открытии порта: ' + config.openHttpPort);
      logger.error(e);
      process.exit(-1);
    }
    else {
      logger.info('Сервер HTTP запущен на порту: ' + config.openHttpPort);
      isServerStart = true;
      tryCheck(agent);
    }
  });
}

// start
try {
  tryMain();
} catch (error) {
  logger.error('Ошибка запуска:');
  logger.error(error);
  process.exit(-1);
}

