const MAX_MONEY_AMOUNT = 99999999.99;

const {
  createValidator,
  oneOf, integer, decimal, minValue, maxValue, required, minLength, maxLength, email, // general validators
  inn, phone, amounts, maxPositionsTotal, maxPaymentsTotal, // specific validators
  or, arrayOf, arrayOfNotBlank, // wrappers
} = require('./index');

const serviceOptions = {
  ca: [required],
  passphrase: [required],
  key: [required],
  cert: [required],
  privateKey: [required],
};

const position = createValidator({
  // количество предмета расчета, 1023: десятичное число с точностью до 6 цифр после запятой, разделитель точка (.)
  quantity: [required, decimal(6)],
  // цена за единицу предмета расчета с учетом скидок и наценок, 1079: десятичное с точностью 2 после запятой, разделитель точка (.)
  price: [required, decimal(2)],
  // ставка НДС, 1199: 20%, 10%, расч. 20/120, расч. 10/110, 0%, не облагается
  tax: [required, integer, oneOf([1, 2, 3, 4, 5, 6])],
  text: [required, maxLength(128)], // наименование предмета расчета, 1030
  // признак способа расчета: предоплата 100%, частичная предоплата, аванс, полный расчет, частичный расч. и кредит,
  // передача в кредит, оплата кредита
  paymentMethodType: [required, oneOf([1, 2, 3, 4, 5, 6, 7])],
  // признак предмета расчета, 1212: 1 товар, 2 подакцизный товар, 3 работа, 4 услуга...
  paymentSubjectType: [required, oneOf([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13])],
  nomenclatureCode: [maxLength(32)], // НЕОБЯЗАТЕЛЬНЫЙ
  supplierINN: [inn], // ИНН поставщика, НЕОБЯЗАТЕЛЬНЫЙ
  supplierInfo: createValidator({ // данные поставщика, // НЕОБЯЗАТЕЛЬНЫЙ
    phoneNumbers: [arrayOf([maxLength(19), phone])],
    name: [maxLength(239)],
  }),
});

const payment = createValidator({
  // 1 - сумма по чеку наличными (1031), 2 - сумма по чеку безналичными (1081), 14 - сумма по чеку предоплатой (зачетом
  // аванса и (или) предыдущих платежей) (1215), 15 - сумма по чеку постоплатой (в кредит) (1216), 16 - сумма по чеку
  // (БСО) встречным предоставлением (1217)
  type: [required, oneOf([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16])],
  amount: [required, decimal(2)], //  сумма оплаты, десятичное с точностью 2 после запятой, разделитель точка (.)
});

const validateRequest = {
  order: createValidator({
    operation_id: [required, integer], // CAT_OPERATIONS.OPERATION_ID
    type: [required, oneOf([1, 2, 3, 4])], // признак расчета, 1054: приход, возврат прихода, расход, возврат расхода
    customerContact: [required, maxLength(64), or([email, phone])], // телефон или электронный адрес покупателя, 1008
  }),
  positions: [required, arrayOfNotBlank(position), maxPositionsTotal(MAX_MONEY_AMOUNT)], // MAX_MONEY_AMOUNT = 99999999.99
  payments: [required, arrayOfNotBlank(payment), maxPaymentsTotal(MAX_MONEY_AMOUNT)]
}

const agent = {
  agentType: [integer, minValue(1), maxValue(127)],
  paymentTransferOperatorPhoneNumbers: [arrayOf([maxLength(19), phone])],
  paymentAgentOperation: [maxLength(24)],
  paymentAgentPhoneNumbers: [arrayOf([maxLength(19), phone])],
  paymentOperatorPhoneNumbers: [arrayOf([maxLength(19), phone])],
  paymentOperatorName: [maxLength(64)],
  paymentOperatorAddress: [maxLength(244)],
  paymentOperatorINN: [inn],
  supplierPhoneNumbers: [arrayOf([maxLength(19), phone])],
};

const isUserAttribute = createValidator({
  name: [required, maxLength(24)],
  value: [required, maxLength(175)],
});

const order = {
  id: [required, maxLength(64)],
  inn: [required, inn],
  group: [maxLength(32)],
  key: [maxLength(32)],
  content: createValidator({
    type: [required, oneOf([1, 2, 3, 4])],
    positions: [required, arrayOf(position), maxPositionsTotal(MAX_MONEY_AMOUNT)],
    checkClose: [amounts, createValidator({
      payments: [required, arrayOf(payment), maxPaymentsTotal(MAX_MONEY_AMOUNT)],
      taxationSystem: [required, integer, oneOf([0, 1, 2, 3, 4, 5])],
    })],
    customerContact: [maxLength(64), or([email, phone])],
    ...agent,
    additionalUserAttribute: [value => value && isUserAttribute(value)],
    automatNumber: [maxLength(20)],
    settlementAddress: [maxLength(243)],
    settlementPlace: [maxLength(243)],
  }),
};
const correction = {
  id: [required, maxLength(64)],
  inn: [required, inn],
  group: [maxLength(32)],
  key: [maxLength(32)],
  content: createValidator({
    type: [required, oneOf([1, 3])],
    taxationSystem: [required, integer, oneOf([0, 1, 2, 3, 4, 5])],
    correctionType: [required, oneOf([0, 1])],
    description: [required, maxLength(243)],
    causeDocumentDate: [required],
    causeDocumentNumber: [required, maxLength(32)],
    totalSum: [required, decimal(2)],
    cashSum: [decimal(2)],
    eCashSum: [decimal(2)],
    prepaymentSum: [decimal(2)],
    postpaymentSum: [decimal(2)],
    otherPaymentTypeSum: [decimal(2)],
    tax1Sum: [decimal(2)],
    tax2Sum: [decimal(2)],
    tax3Sum: [decimal(2)],
    tax4Sum: [decimal(2)],
    tax5Sum: [decimal(2)],
    tax6Sum: [decimal(2)],
    automatNumber: [maxLength(20)],
    settlementAddress: [maxLength(243)],
    settlementPlace: [maxLength(243)],
  }),
};

module.exports.serviceOptions = serviceOptions;
module.exports.order = order;
module.exports.correction = correction;
module.exports.validateRequest = validateRequest;